# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""GNU/Linux distributions defines."""

import grp
from os.path import exists
from gettext import gettext as _

from libmace import lsb_release, util


#: Dictionary to map unified package names to exact ones
PACKAGES = {
        'apache': {
            ('Fedora', None): 'httpd',
            },
        'openssh': {
            ('Fedora', None): 'openssh-server',
            },
        'squidguard': {
            ('Fedora', None): 'squidGuard',
            },
        }


def is_daemon(name):
    """Is specified service registered in the system as a daemon.

    :param name:
        the name of init.d script
    :returns:
        `True` if service installed

    """
    return exists('/etc/init.d/%s' % name)


def call_daemon(name, command, **kwargs):
    """Execute daemon command.

    :param name:
        the name of init.d script
    :param command:
        command to execute, like ``status`` or ``start``
    :param kwargs:
        additional arguments to pass to `util.call()`
    :returns:
        `util.call()` output

    """
    return util.call(['/etc/init.d/%s' % name, command], **kwargs)


def resolve_package(name):
    """Check if specified package installed.

    :param name:
        unified package name, the exact name to current distribution
        will be taken from `PACKAGES` dictionary
    :returns:
        `True` if package is installed

    """
    lsb_id = lsb_release.distributor_id()
    lsb_rel = lsb_release.release()

    if name in PACKAGES:
        for (distro_name, distro_release), mapping in PACKAGES[name].items():
            if distro_name == lsb_id and (distro_release is None or \
                    distro_release == lsb_rel):
                name = mapping
                break
        else:
            raise RuntimeError(
                    _('Cannot find package names mapping for %s on %s-%s') % \
                            (name, lsb_id, lsb_rel))

    if lsb_id == 'Fedora':
        installed = util.call(['rpm', '-q', name]) is not None
    else:
        raise RuntimeError(_('Not supported distribution, %s') % lsb_id)

    return name, installed


def group_add(name):
    util.assert_call(['/usr/sbin/groupadd', name])


def user_add(name, group=None, managehome=None, groups=None):
    args = ['/usr/sbin/useradd']
    if group:
        args.extend(['-g', group])
    if managehome:
        args.extend(['-m'])
    if groups:
        args.extend(['-G', ','.join(groups)])
    util.assert_call(args + [name])


def user_modify(name, group=None, home=None, groups=None):
    args = []
    if group:
        args.extend(['-g', group])
    if home:
        args.extend(['-d', home])
    if groups:
        args.extend(['-G', ','.join(groups)])
    if args:
        util.assert_call(['/usr/sbin/usermod'] + args + [name])


def get_user_groups(user):
    groups = set()
    for i in grp.getgrall():
        if user in i.gr_mem:
            groups.add(i.gr_name)
    return groups
