# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Process configuration files."""

import os
import re
import pwd
import grp
import shlex
import shutil
import logging
from glob import glob
from cStringIO import StringIO
from ConfigParser import ConfigParser
from os.path import join, dirname, basename, relpath, abspath
from os.path import isfile, isdir, expandvars, exists, islink
from gettext import gettext as _

from libmace import env, printf, util, distro
from libmace.util import enforce


NOTICE = """\
WARNING: File was generate by mace utility, do not edit it.
         Change files in %s directory instead and run mace."""

LIST_SEPARATOR = ';'


def load(src_dir, providers):
    """Load tree of configuration sources.

    :param src_dir:
        path to directories tree with configuration sources
    :param providers:
        list of `libmace.providers.BaseProvider` objects to scan
        configuration for
    :returns:
        a tuple of ({provider: `Conftext`}, {var_name: var_value});
        the first item is for found sources and the last one is for found
        variables

    """
    found = {}
    skipped = set()
    unknown = set()
    src_dir = abspath(src_dir)

    def load_files(path):
        provider = None
        for provider in providers:
            if ('%s.d' % provider.name) == basename(path):
                break
        else:
            unknown.add(basename(path))
            return

        if env.include.value and \
                not _match_service(env.include.value, provider.name):
            skipped.add(provider.name)
            return
        if env.exclude.value and \
                _match_service(env.exclude.value, provider.name):
            skipped.add(provider.name)
            return

        conftext = found.get(provider) or Conftext()
        for conf_prio, conf_name in enumerate(provider.files):
            for i in glob(join(path, conf_name or '', '*.conf')):
                keywords = re.findall('\[([^]]+)\]', i)
                if not keywords or set(keywords) & set(env.keywords.value):
                    conftext.append(SourceFile(
                            provider, src_dir, conf_name, conf_prio, i))

        if conftext or not provider.files:
            conftext.sort()
            found[provider] = conftext

    envar_paths = []
    if src_dir.endswith('.d'):
        load_files(src_dir)
    else:
        for root, dirs, files in os.walk(src_dir, followlinks=True):
            dirs.sort()
            for entry in sorted(dirs[:] + files):
                entry_path = join(root, entry)
                if entry.endswith('.env'):
                    if isfile(entry_path):
                        envar_paths.append(entry_path)
                elif entry.endswith('.d'):
                    if isdir(entry_path):
                        load_files(entry_path)
                        dirs.remove(entry)

    for confdir in found.values():
        for i, f in enumerate(confdir):
            for j in range(i):
                confdir[j].override(f)

    envars = {}
    for provider, __ in found.items():
        provider.export(envars)
    for key, value in [
            ('SRC_ROOT', env.source.value),
            ('DST_ROOT', env.output.value)]:
        envars[key] = value
    for key, value in envars.items():
        os.environ[key] = value
    for i in envar_paths:
        envars.update(_parse_env(i))

    if skipped:
        logging.info(_('Skip %s due to --include/--exclude arguments'),
                ', '.join(skipped))
    if unknown:
        logging.warning(_('No providers for %s'), ', '.join(unknown))

    return found, envars


def commit(src_dir, dst_dir, sources, envars, force=False, dry_services=False):
    """Apply configuration to final environment.

    :param src_dir:
        path to directories tree with configuration sources
    :param dst_dir:
        path to directories tree to write configuration to
    :param sources:
        dictionary returned by `load()` method
    :param force:
        force to rewrite configuration and reapply it
    :param dry_services:
        do not ask providers to apply configuration,
        just write configuration files
    :returns:
        list of providers that configuration was applied

    """
    enforce(sources, _('Cannot find any sources'))
    logging.info(_('Found sources for: %s'),
            ', '.join([i.name for i in sources.keys()]))

    providers = []
    for i in sources.keys():
        providers.extend(i.provides)

    for i in sources.keys():
        enforce(not (set(i.depends) - set(providers)),
                _('Provider %s needs %r services to be included ' \
                        'to the configuration sources'), i.name, i.depends)

    Transaction.envars = envars
    applied = []

    for provider, conftext in sources.items():
        if not dry_services:
            missed_packages = []
            for package in provider.get_requires(conftext):
                package, installed = distro.resolve_package(package)
                if not installed:
                    missed_packages.append(package)
            if missed_packages:
                logging.warning(_('Service %s is not installed, ' \
                        'no %s package(s)'), provider.name,
                        ' '.join(missed_packages))
                continue
            enforce(provider.service_name is None or \
                    distro.is_daemon(provider.service_name),
                    _('Service %s is installed but cannot find %s ' \
                            'init.d script'),
                    provider.name, provider.service_name)

        if not dry_services and provider.enable():
            logging.info(_('Registered %s service to start at boot time'),
                    provider.name)

        if provider.comment:
            header = ['%s %s' % (provider.comment, i) \
                    for i in (NOTICE % src_dir).split('\n')]
            header.append('')
        else:
            header = []

        configs = {}
        for conf_name in provider.files:
            if not conftext.confile(conf_name):
                continue
            confile = provider.confile(conf_name)
            if confile not in configs:
                configs[confile] = []
            configs[confile].extend(
                    provider.confile_header(conf_name, conftext))
            for i in conftext.confile(conf_name):
                configs[confile].extend(i.content)
            configs[confile].extend(
                    provider.confile_footer(conf_name, conftext))

        tran = Transaction(dst_dir)
        try:
            for confile, lines in configs.items():
                if confile is None:
                    tran.config.read_and_expand(lines, envars)
                else:
                    lines = [_expand(i, envars) for i in lines]
                    lines = provider.convert(lines)
                    # The last NL is imprtant, e.g., iptables-restore needs it
                    # for the last COMMIT directive
                    config = '\n'.join(header + lines) + '\n'
                    tran.ensure_file(confile, content=config)

            if dry_services:
                if tran.config.sections():
                    printf.info(_('Silently apply configuration for %s'),
                            provider.name)
                continue

            provider.ensure(tran)

            if tran.config.sections():
                if provider.apply(tran):
                    printf.info(_('Apply configuration for %s'), provider.name)
                applied.append(provider)
            elif force or not tran.empty:
                printf.info(_('Apply configuration for %s'), provider.name)
                provider.apply(tran)
                applied.append(provider)
            else:
                logging.info(('Configuration for %s was not changed'),
                        provider.name)
                if provider.activate():
                    printf.info(('Service %s is started'), provider.name)
        except Exception:
            printf.exception(_('Cannot apply %s configuration, rollback'),
                    provider.name)
            tran.rollback()

    return applied


def stat(sources, cb):
    """Return configuration sources statistic.

    :param sources:
        dictionary returned by `load()` function
    :param cb:
        callback that will be called with `SourceFile` object and should
        return `True` if this file needs to be included to result statistic
    :returns:
        a tuple of (total, merged, overriden, hidden) numbers of files

    """
    total = 0
    rel_paths = set()
    overriden = 0
    hidden = 0

    for __, confiles in sources.items():
        for i in confiles:
            if cb(i):
                total += 1
                if i.hidden:
                    hidden += 1
                if i.overriden:
                    overriden += 1
                rel_paths.add(i.rel_path)

    if not total:
        return 0, None, None, None
    else:
        merged = len(set([basename(i) for i in rel_paths])) < len(rel_paths)
        return total, merged, overriden, hidden


def uid_by_name(name):
    """Get uid by user name.

    Should be used instead of `pwd.getpwnam`. Was introduced only to support
    unittests.

    :param name:
        user name
    :returns:
        uid integer value

    """
    return pwd.getpwnam(name).pw_uid


def gid_by_name(name):
    """Get gid by group name.

    Should be used instead of `grp.getgrnam`. Was introduced only to support
    unittests.

    :param name:
        user name
    :returns:
        uid integer value

    """
    return grp.getgrnam(name).gr_gid


def chown(path, uid, gid):
    """Change file ownership.

    Should be used instead of `os.chown`. Was introduced only to support
    unittests.

    :param uid:
        user id
    :param gid:
        group id

    """
    os.chown(path, uid, gid)


class Conftext(list):
    """Collect configuration sources for the same provider."""

    def confile(self, confile):
        """`SourceFile` objects for the same confile.

        :param confile:
            the name of confile
        :returns:
            list of `SourceFile` objects

        """
        return [i for i in self if i.confile == confile]


class SourceFile(object):
    """Source file representation.

    This class describes only source configuration file, i.e., it is a part of
    source confile and there are might be several files for the same confile
    in sources tree.

    """

    def __init__(self, provider, src_dir, confile, confile_prio, path):
        """
        :param provider:
            `libmace.providers.BaseProvider` object that this configuration
            file belongs to
        :param src_dir:
            the sources root directory
        :param confile:
            provider's confile name
        :param confile_prio:
            confile priority; less value means higher priority
        :param path:
            path to this source file

        """
        self.content = []
        self.hidden = False
        self.overriden = False
        self.hidden_by = set()
        self.overriden_by = set()

        self.provider = provider
        self.rel_path = dirname(relpath(path, src_dir))
        if confile:
            self.rel_path = dirname(self.rel_path)
            self.filename = join(confile, basename(path))
        else:
            self.filename = basename(path)
        self.path = path
        self.confile = confile
        self.confile_prio = confile_prio

        for line in file(path).readlines():
            self.content.append(line.rstrip('\n'))

    def override(self, other):
        """Check if this configuration file is being overriden by another one.

        :param other:
            `SourceFile` that substitutes this one

        """
        if self.provider is other.provider and self.filename == other.filename:
            if _is_empty(other.content, self.provider.comment):
                self.hidden = True
                self.hidden_by.add(other.rel_path)
            else:
                self.overriden = True
                self.overriden_by.add(other.rel_path)

    def __cmp__(self, other):
        if other is None:
            return cmp(self.filename, None)
        r = cmp((self.confile_prio, self.filename),
                (other.confile_prio, other.filename))
        if r != 0:
            return r
        return cmp(self.rel_path, other.rel_path)


class Transaction(object):
    """Configuration application transation.

    These objects will be used while applying final configuration. All changes
    need to be done via `Transaction` methods to let it rollback changes
    on fails.

    """
    #: Variables exported by all providers
    envars = {}

    def __init__(self, root):
        """
        :param root:
            path to the root directory with final configuration, e.g., ``/``

        """
        self._root = root
        self._backup = {}

        #: Artificial provider configuration.
        #: Some providers might support artificial configuration, i.e.,
        #: if there is no singular configuration syntax. For these providers
        #: `config` will contain configuration parsed from `.ini` files.
        self.config = Config()

    @property
    def empty(self):
        """Are there any changes made within this transation."""
        return not self._backup

    def path(self, path):
        """Arbitrary path starting from final configuration root directory.

        :param path:
            relative path
        :returns:
            specified `path` with prefixing root directory

        """
        if path.startswith(self._root):
            return path
        else:
            return join(self._root, path.lstrip(os.sep))

    def ensure_file(self, path, content=None, owner=None, mode=None,
            create_missed=None):
        """Change a file.

        :param path:
            configuration file within the root directory to track
        :param content:
            if not `None`, content that `path` file should have
        :param owner:
            what the owner of file should be; the singular value will define
            user and group, but it might be a tuple; atomic values might be
            a string to define names or ints to define ids
        :param mode:
            mode that file should have
        :param create_missed:
            create files if they don't exist
        :returns:
            `True` if something was changed

        """
        stale = False
        if isinstance(content, list) or isinstance(content, tuple):
            content = '\n'.join([i.rstrip('\n') for i in content])
        path = self.path(path)

        if exists(path):
            if content is not None and content != file(path).read():
                self._keep_file(path, content is not None)
        elif create_missed:
            self._keep_file(path, True)
            if not exists(dirname(path)):
                os.makedirs(dirname(path))
            file(path, 'w').close()
        else:
            enforce(content is not None,
                    _('Configuration file %s is absent'), path)
            self._backup[path] = None
            if not exists(dirname(path)):
                os.makedirs(dirname(path))

        if content is not None:
            f = file(path, 'w')
            f.write(content)
            f.close()

        if owner:
            uid, gid = _get_cred_ids(owner)
            stats = os.stat(path)
            if stats.st_uid != uid or stats.st_gid != gid:
                self._keep_file(path)
                chown(path, uid, gid)
                stale = True

        if mode is not None and (os.stat(path).st_mode & 0777) != mode:
            self._keep_file(path)
            os.chmod(path, mode)
            stale = True

        return stale

    def ensure_dir(self, path, owner=None, mode=None):
        """Change a directory.

        :param path:
            configuration file within the root directory to track
        :param owner:
            what the owner of directory should be; the singular value will
            define user and group, but it might be a tuple; atomic values
            might be a string to define names or ints to define ids
        :param mode:
            mode that directory should have
        :returns:
            `True` if something was changed

        """
        stale = False
        path = self.path(path)

        if path in self._backup:
            enforce(not isinstance(self._backup[path], util.TempDir),
                    _('Cannot overwrite already configured file '
                    'by new directory'))
            return

        if exists(path):
            if islink(path) or not isdir(path):
                self._keep_file(path, True)
                os.makedirs(path)
                stale = True
        else:
            self._backup[path] = None
            os.makedirs(path)
            stale = True

        def keep():
            if path not in self._backup:
                self._backup[path] = os.stat(path)

        if owner:
            uid, gid = _get_cred_ids(owner)
            stats = os.stat(path)
            if stats.st_uid != uid or stats.st_gid != gid:
                keep()
                chown(path, uid, gid)
                stale = True

        if mode is not None and (os.stat(path).st_mode & 0777) != mode:
            keep()
            os.chmod(path, mode)
            stale = True

        return stale

    def ensure_symlink(self, src, dst):
        """Create a symlink.

        :param src:
            configuration file within the root directory to link from
        :param dst:
            configuration file within the root directory to link to
        :returns:
            `True` if something was changed

        """
        dst = self.path(dst)

        if islink(dst):
            if os.readlink(dst) == src:
                return False
            os.unlink(dst)
        elif exists(dst):
            os.unlink(dst)

        os.symlink(src, dst)
        return True

    def rollback(self):
        """Revert all previsly made changes within this transation."""
        while self._backup:
            path, backup = self._backup.popitem()

            if exists(path):
                if islink(path):
                    os.unlink(path)
                elif not isdir(path):
                    os.unlink(path)

            if isinstance(backup, util.TempDir):
                shutil.rmtree(path, ignore_errors=True)
                os.rename(join(backup, basename(path)), path)
            elif backup is None:
                if exists(path):
                    if isdir(path):
                        shutil.rmtree(path)
                    else:
                        os.unlink(path)
            else:
                os.chmod(path, backup.st_mode)
                chown(path, backup.st_uid, backup.st_gid)

    def checkpoint(self):
        """Cleanup rollback history, so `rollback()` call will do nothing."""
        self._backup.clear()

    def _keep_file(self, path, content_change=False):
        if path in self._backup:
            if content_change:
                # Since backup and original files might be hard linked,
                # remove original file before writing to it
                os.unlink(path)
            return
        elif exists(path):
            tmp_dir = util.TempDir(prefix=basename(path), dir=dirname(path))
            tmp_path = join(tmp_dir, basename(path))
            shutil.copy2(path, tmp_path)
            stats = os.stat(path)
            chown(tmp_path, stats.st_uid, stats.st_gid)
            self._backup[path] = tmp_dir
        else:
            self._backup[path] = None


class Config(ConfigParser):

    def read_and_expand(self, lines, envars):
        content = StringIO('\n'.join(lines))

        # Expanding raw text before reading it by ConfigParser might be
        # problematic becase expanded option values might be multilined
        # thus, do expanding after reading to the temporary config
        tmp = ConfigParser()
        tmp.readfp(content)

        for section in tmp.sections():
            section_ = _expand(section, envars)
            self.add_section(section_)
            for option, value in tmp.items(section):
                self.set(section_,
                        _expand(option, envars), _expand(value, envars))

    def get_list(self, section, key):
        str_list = self.get(section, key)
        if not str_list:
            return []

        parts = []
        brackets = {('(', ')'): 0,
                    ('[', ']'): 0,
                    ('"', '"'): 0}
        str_list = str_list.replace("\n", LIST_SEPARATOR).strip()
        i = 0

        while i < len(str_list):
            if not max(brackets.values()) and str_list[i] == LIST_SEPARATOR:
                parts.append(str_list[:i])
                str_list = str_list[i + 1:]
                i = 0
            else:
                for key in brackets.keys():
                    left, right = key
                    if str_list[i] == left:
                        brackets[key] += 1
                        break
                    elif str_list[i] == right:
                        brackets[key] -= 1
                        break
                i += 1

        parts.append(str_list)

        return [i.strip() for i in parts if i.strip()]

    def __getitem__(self, key):
        if isinstance(key, tuple):
            section, key = key
        else:
            section = 'DEFAULT'
        if self.has_option(section, key):
            return self.get(section, key)
        else:
            pass


def _parse_env(path):
    result = {}

    for token in shlex.split(file(path).read(), comments=True):
        if '=' not in token:
            logging.warning(_('Cannot parse variable value from line ' \
                    '"%s" in %s'), token, path)
        key, value = token.split('=', 1)
        result[key] = expandvars(value)
        os.environ[key] = result[key]

    return result


def _expand(line, envars):
    while True:
        for key, value in envars.items():
            new_line = line.replace('@%s@' % key, value)
            if new_line != line:
                line = new_line
                break
        else:
            break
    return line


def _is_empty(content, comment):
    return not [i for i in content if not i.startswith(comment) and i.strip()]


def _match_service(array, name):
    return name in array or '%s.d' % name in array


def _get_cred_ids(creds):
    if isinstance(creds, tuple) or isinstance(creds, list):
        user, group = creds
        group = group or user
    else:
        user = group = creds
    if not isinstance(user, int):
        user = uid_by_name(user)
    if not isinstance(group, int):
        group = gid_by_name(group)
    return user, group
