# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Providers related defines."""

import os
import imp
from os.path import join, dirname, abspath
from gettext import gettext as _

from libmace import util, env, printf, distro


PROVIDERS_DIR = join(abspath(dirname(__file__)), 'providers')


class BaseProvider(object):
    """All providers need to inherit this class."""

    @property
    def comment(self):
        """String that is being used to start comment line in config files."""
        raise NotImplementedError()

    @property
    def name(self):
        """Provider identification string."""
        raise NotImplementedError()

    @property
    def provides(self):
        """List of names that are associated with a provider.

        By default providers are associated only with theirs names.

        """
        return [self.name]

    @property
    def depends(self):
        """Providers that current one requires.

        All depending providers need to be included to the same configuration
        as current one belongs to. It is only about depends between providers
        not between services they configure, e.g., some provider might depend
        on envars exported by enother provider.

        """
        return []

    @property
    def service_name(self):
        """Name of init.d service associated with the provider."""
        raise NotImplementedError()

    @property
    def files(self):
        """Configuration files supported by provider.

        Property contains fake names internally associated with final
        configuration files. If there is only one configuration file
        supported by provider, names don't make sense and even None is fine.

        """
        raise NotImplementedError()

    def confile(self, filename):
        """Get configuration path relatively to --output directory.

        :param filename:
            one of `files` filenames
        :returns:
            path to the final configuration file associated with `filename`

        """
        raise NotImplementedError()

    def confile_header(self, filename, conftext):
        """The header that will be added to each final configuration file.

        :param filename:
            one of `files` filenames
        :param conftext:
            `libmace.files.Conftext` object with full configuration
        :returns:
            list of lines to add

        """
        return []

    def confile_footer(self, filename, conftext):
        """The footer that will be added to each final configuration file.

        :param filename:
            one of `files` filenames
        :param conftext:
            `libmace.files.Conftext` object with full configuration
        :returns:
            list of lines to add

        """
        return []

    def confpath(self, filename=None):
        """Get full configuration path, including --output directory.

        :param filename:
            one of `files` filenames
        :returns:
            path for real configuration file

        """
        path = self.confile(filename)
        if path.startswith(env.output):
            return path
        else:
            return join(env.output, path.lstrip(os.sep))

    def enable(self):
        """Make sure that service will start at boot time.

        :returns:
            `True` is service was added to start list

        """
        if self.service_name is None:
            return
        if not self._is_service_registered():
            self._register_service()
            return True

    def activate(self):
        """Is service running.

        :returns:
            `True` is service was started

        """
        if self.service_name is None:
            return
        if distro.call_daemon(self.service_name, 'status') is None:
            distro.call_daemon(self.service_name, 'start', asserts=True)
            return True

    def export(self, envars):
        """Export variables that might be used in configuration sources.

        :param envars:
            dictionary to export variables to

        """
        pass

    def convert(self, lines):
        """Convert configuration file.

        :param lines:
            configuration file text lines
        :returns:
            function should return new lines

        """
        return lines

    def ensure(self, tran):
        """Does environment needs to be prepared.

        Will be called after applying new configuration to make sure that
        final environment for consistence and tunning it otherwise.

        :param tran:
            any changes in environment should happen via this object
            to make it possible to revert made changes on applying fails
        :returns:
            `True` if environment was changed

        """
        pass

    def apply(self, tran):
        """Apply configuration changes, e.g., restart service.

        :param tran:
            any changes in environment should happen via this object
            to make it possible to revert made changes on applying fails
        :returns:
            `True` if provider has done something for applying changes, e.g.,
            restarted daemon

        """
        if self.service_name is not None:
            distro.call_daemon(self.service_name, 'restart', asserts=True)
            return True

    def get_requires(self, conftext):
        """Packages that provider requires.

        :param conftext:
            `libmace.files.Conftext` object with full configuration
        :returns:
            list of package aliases that will be resolved to particular
            distribution names using `libmace.distro.PACKAGES` dictionary

        """
        return []

    def _is_service_registered(self):
        return util.call(['chkconfig', self.service_name]) is not None

    def _register_service(self):
        return util.assert_call(['chkconfig', self.service_name, 'on'])


def load():
    """Populate list of providers.

    Needs to be called on program start.

    """
    result = []

    for mod_file in os.listdir(PROVIDERS_DIR):
        if not mod_file.endswith('.py'):
            continue
        mod_name = mod_file[:-3]
        fp, pathname, description = imp.find_module(mod_name, [PROVIDERS_DIR])
        try:
            mod = imp.load_module(mod_name, fp, pathname, description)
            provider = mod.Provider()
            result.append(provider)
        except Exception:
            printf.exception(_('Cannot load provider from %s file'),
                    join(PROVIDERS_DIR, mod_file))
        finally:
            if fp:
                fp.close()

    return result
