# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Apache
------

Provider will do some extra work:

- comment out all LoadModule directives in main config file to let mace
  configuration load only particular modules
- comment out all ErrorLog and CustomLog directives in main config file
  to let mace configuration full control on logs
- make sure that log directories are properly created

Mace configuration:
    :Service:
        ``httpd.d``
    :Confiles:
        - ``modules`` only for `LoadModule` directives
        - ``conf`` the rest of configuration except vhosts
        - ``vhosts`` convenient confile to collect all vhosts declarations
    :Variables:
        - `HTTP_USER` system user http server uses
        - `HTTP_GROUP` system group http server uses

Fedora Final configuration:
    :Replaces:
        Partially, provider will try to remove `LoadModule` directives from
        basic configuration to let Mace configuration full control on what
        modules are enabled; the rest of configuration will happen in
        `conf.d` style
    :Files:
        - ``/etc/httpd/conf.d/000-mace.conf`` to keep `LoadModule` at the top
        - ``/etc/httpd/conf.d/zzz-mace.conf`` the rest of declarations

"""
import re
import logging
from os.path import exists
from gettext import gettext as _

from libmace.providers import BaseProvider


_LOG_PARAMETER_RE = re.compile('\s*(ErrorLog|CustomLog)\s+([^\s]+)')


class Provider(BaseProvider):

    @property
    def comment(self):
        return '#'

    @property
    def name(self):
        return 'apache'

    @property
    def provides(self):
        return [self.name, 'httpd']

    def get_requires(self, conftext):
        return ['apache']

    @property
    def service_name(self):
        return 'httpd'

    @property
    def files(self):
        return ['modules', 'conf', 'vhosts']

    def confile(self, filename=None):
        if filename == 'modules':
            return '/etc/httpd/conf.d/000-mace.conf'
        else:
            return '/etc/httpd/conf.d/zzz-mace.conf'

    def export(self, envars):
        envars['HTTP_USER'], envars['HTTP_GROUP'] = self._owner

    def ensure(self, tran):
        self._ensure_log_paths(tran)
        self._tweak_master_conf(tran)
        self._hide_proxy_ajp(tran)

    @property
    def _owner(self):
        return ('apache', 'apache')

    def _ensure_log_paths(self, tran):
        for line in file(tran.path(self.confile())).readlines():
            match = _LOG_PARAMETER_RE.match(line)
            if match is None:
                continue
            logpath = tran.path(match.group(2))
            tran.ensure_file(logpath, owner=self._owner, create_missed=True)

    def _tweak_master_conf(self, tran):
        """Disable LoadModule directives.

        We need full control on what modules will be loaded, e.g.,
        to minimize memory usage. All used modules need to be enabled
        explicitly in mace configuration sources.

        """
        master_conf = tran.path('/etc/httpd/conf/httpd.conf')
        if not exists(master_conf):
            return

        new_conf = []
        modified = False

        for line in file(master_conf).readlines():
            line = line.strip()
            for i in ['LoadModule', 'ErrorLog', 'CustomLog']:
                if line.startswith(i):
                    line = '# ' + line
                    modified = True
                    break
            new_conf.append(line)

        if modified:
            logging.info(_('Tweak %s to let mace configuration full control'),
                    master_conf)
            tran.ensure_file(master_conf, content=new_conf)

    def _hide_proxy_ajp(self, tran):
        """Avoid loading proxy_ajp because mod_proxy might be disabled.

        At least httpd from Fedora-11 enables proxy_ajp out-of-the-box.
        It doesn't work since all modules are disabled by default and mod_proxy
        might be not enabled in mace configuration sources.

        """
        conf_path = tran.path('/etc/httpd/conf.d/proxy_ajp.conf')
        if not exists(conf_path):
            return

        conf = [i.strip() for i in file(conf_path).readlines()]
        if not [i for i in conf \
                if i.startswith('LoadModule proxy_ajp_module')] or \
                conf[0] == '<IfModule mod_proxy>' and \
                conf[-1] == '</IfModule>':
            return

        logging.info(_('Load proxy_ajp_module only if mod_proxy is loaded'))

        tran.ensure_file(conf_path,
                content=['<IfModule mod_proxy>'] + conf + ['</IfModule>'])
