# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
PAFM
----

`PHP AJAX File Manager <http://sourceforge.net/projects/pafm>`_
`patched <http://git.sugarlabs.org/server/pafm>`_ to have configuration file in
``/etc/pafm.conf``.

Provider will do some extra work:

- set httpd user as an owner for ``/etc/pafm.conf``
- set 0600 permissions for ``/etc/pafm.conf`` to hide password for others
- ensure that ``files_root`` directory specified in configuration file created
- symlink php files adn ``files_root`` to the ``www_root`` directory

Mace configuration:
    :Service:
        ``pafm.d``
    :Confiles:
        No
    :Variables:
        No

Fedora final configuration:
    :Replaces:
        Yes
    :Files:
        ``/etc/pafm.conf``

"""
from os.path import join, isabs
from ConfigParser import ConfigParser
from gettext import gettext as _

from libmace.providers import BaseProvider
from libmace.util import enforce


_INSTDIR = '/usr/share/pafm'


class Provider(BaseProvider):

    @property
    def comment(self):
        return '#'

    @property
    def name(self):
        return 'pafm'

    @property
    def depends(self):
        return ['httpd']

    def get_requires(self, conftext):
        return ['pafm']

    @property
    def service_name(self):
        return None

    @property
    def files(self):
        return [None]

    def confile(self, filename=None):
        return '/etc/pafm.conf'

    def ensure(self, tran):
        enforce('HTTP_USER' in tran.envars,
                _('HTTP service should export HTTP_USER variable'))
        enforce('HTTP_GROUP' in tran.envars,
                _('HTTP service should export HTTP_GROUP variable'))

        owner = (tran.envars['HTTP_USER'], tran.envars['HTTP_GROUP'])

        # Configuration file contains plain password
        tran.ensure_file(self.confile(), owner=owner, mode=0600)

        config = ConfigParser()
        config.read(tran.path(self.confile()))

        enforce(config.has_option('main', 'www_root'),
                _('PAFM configuration does not contain main/www_root'))
        www_root = config.get('main', 'www_root')
        enforce(isabs(www_root),
                _('PAFM main/www_root directory should be absolute'))
        tran.ensure_dir(www_root, owner=owner, mode=0755)
        tran.ensure_symlink(
                join(_INSTDIR, 'pafm.php'), join(www_root, 'index.php'))
        tran.ensure_symlink(join(_INSTDIR, 'pafm'), join(www_root, 'pafm'))

        enforce(config.has_option('main', 'files_root'),
                _('PAFM configuration does not contain main/files_root'))
        files_root = config.get('main', 'files_root')
        enforce(isabs(files_root),
                _('PAFM main/files_root directory should be absolute'))
        tran.ensure_dir(files_root, owner=owner, mode=0755)
        tran.ensure_symlink(files_root, join(www_root, 'files'))
