# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Prosody
-------

Provider will do some extra work:

- make sure that data directory is properly created
- make sure that log directory is properly created
- make sure that server self-signed SSL cerificate is good
  and recreate it oterwise

Mace configuration:
    :Service:
        ``prosody.d``
    :Confiles:
        No
    :Variables:
        No

Fedora final configuration:
    :Replaces:
        Yes
    :Files:
        ``/etc/prosody/prosody.cfg.lua``

"""

import re
from os.path import exists, dirname
from gettext import gettext as _

from libmace import util
from libmace.providers import BaseProvider
from libmace.util import enforce


_CREDS = ('prosody', 'prosody')

_OPENSSL = '/usr/bin/openssl'
_OPENSSL_KEY_SIZE = 2048
_OPENSSL_CERT_DAYS = 365

_SSL_RE = re.compile('^\s*ssl\s*=\s*{([^}]*)}', re.MULTILINE)
_LOG_RE = re.compile('^\s*log\s*=\s*{([^}]*)}', re.MULTILINE)
_OPT_RE = re.compile('([a-z]+)\s*=\s*"([^"]*)"')
_POSIX_RE = re.compile('^\s*modules_enabled[^}]*^\s*"posix"', re.MULTILINE)

OPENSSL_CONF = """
oid_section = new_oids

[ new_oids ]

# RFC 3920 section 5.1.1 defines this OID
xmppAddr = 1.3.6.1.5.5.7.8.5

# RFC 4985 defines this OID
SRVName  = 1.3.6.1.5.5.7.8.7

[ req ]

default_bits       = 4096
default_keyfile    = %(vhost)s.key
distinguished_name = distinguished_name
req_extensions     = v3_extensions
x509_extensions    = v3_extensions

# ask about the DN?
prompt = no

[ distinguished_name ]

commonName             = %(vhost)s
countryName            = RU
localityName           = The Internet
organizationName       = Your Organisation
organizationalUnitName = XMPP Department
emailAddress           = xmpp@%(vhost)s

[ v3_extensions ]

# for certificate requests (req_extensions)
# and self-signed certificates (x509_extensions)

basicConstraints = CA:FALSE
keyUsage         = digitalSignature,keyEncipherment
extendedKeyUsage = serverAuth,clientAuth
subjectAltName   = @subject_alternative_name

[ subject_alternative_name ]

DNS.0       =                                           %(vhost)s
otherName.0 =                 xmppAddr;FORMAT:UTF8,UTF8:%(vhost)s
otherName.1 =            SRVName;IA5STRING:_xmpp-client.%(vhost)s
otherName.2 =            SRVName;IA5STRING:_xmpp-server.%(vhost)s

DNS.1       =                                conference.%(vhost)s
otherName.3 =      xmppAddr;FORMAT:UTF8,UTF8:conference.%(vhost)s
otherName.4 = SRVName;IA5STRING:_xmpp-server.conference.%(vhost)s
"""


class Provider(BaseProvider):

    @property
    def comment(self):
        return '--'

    @property
    def name(self):
        return 'prosody'

    def get_requires(self, conftext):
        return ['prosody-sugar']

    @property
    def service_name(self):
        return 'prosody'

    @property
    def files(self):
        return [None]

    def confile(self, filename=None):
        return '/etc/prosody/prosody.cfg.lua'

    def confile_header(self, filename, conftext):
        return ['pidfile = "/var/run/prosody.pid"',
                'run_as_root = false',
                ]

    def ensure(self, tran):
        config = file(tran.path(self.confile())).read()

        enforce(_POSIX_RE.search(config) is not None, \
                _('Prosody config should include "posix" plugin to enable ' \
                'pidfile support'))

        data_path = _parse_value('data_path', config)
        enforce(data_path, _('Parameter data_path should be set'))
        tran.ensure_dir(data_path, owner=_CREDS)

        log_path = _parse_value('log', config)
        if log_path is not None:
            log_paths = [log_path]
        else:
            log_paths = (_parse_struct(_LOG_RE, config) or {}).values()
        for i in log_paths:
            tran.ensure_dir(dirname(i), owner=_CREDS)

        _ensure_ssl_certs(tran, config)

        return tran


def _ensure_ssl_certs(tran, config):
    newkey = None

    ssl = _parse_struct(_SSL_RE, config)
    if ssl is None:
        return

    enforce(ssl.get('key') and ssl.get('certificate'),
            _('Options key and certificate were not specified for ssl'))

    key_path = tran.path(ssl['key'])
    cert_path = tran.path(ssl['certificate'])

    for i in [key_path, cert_path]:
        tran.ensure_dir(dirname(i), owner=_CREDS, mode=0700)

    if not exists(key_path):
        newkey = util.assert_call([_OPENSSL, 'genrsa', str(_OPENSSL_KEY_SIZE)])
        tran.ensure_file(key_path, content=newkey)
    tran.ensure_file(key_path, mode=0400, owner=_CREDS)

    if newkey or not exists(cert_path) or not _verify_cert(cert_path):
        vhost = _parse_directive_value('VirtualHost', config)
        enforce(vhost, _('No VirtualHost directives in config'))
        openssl_conf = util.TempFilePath(text=OPENSSL_CONF % {'vhost': vhost})
        tran.ensure_file(cert_path, content=[])
        util.assert_call([_OPENSSL, 'req', '-new', '-x509', '-nodes', '-sha1',
                '-utf8', '-key', key_path, '-days', str(_OPENSSL_CERT_DAYS),
                '-out', cert_path, '-config', openssl_conf])
    tran.ensure_file(cert_path, owner=_CREDS)


def _verify_cert(path):
    verified = util.call([_OPENSSL, 'verify', path])
    return verified is not None and \
            not [i for i in verified.split('\n') if i.startswith('error 10 ')]


def _parse_value(name, text):
    match = re.search('^\s*%s\s*=\s*"([^"]*)"' % name, text, re.MULTILINE)
    return None if match is None else match.group(1)


def _parse_directive_value(name, text):
    match = re.search('^\s*%s\s+"([^"]*)"' % name, text, re.MULTILINE)
    return None if match is None else match.group(1)


def _parse_struct(regexp, text):
    match = regexp.search(text)
    if match is None:
        return None

    result = {}

    for option in match.group(1).split(';'):
        match = _OPT_RE.match(option.strip())
        if match is None:
            continue
        opt, value = match.groups()
        result[opt] = value

    return result
