# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Sarg
----

`Squid Analysis Report Generator <http://sarg.sourceforge.net/sarg.php>`_.

Provider will do some extra work:

- `language` configuration option should be in ISO-639-1 notation that
  will be replaced to language names that Sarg supports
- make sure that `output_dir` is created and have 0755 access permissions

Mace configuration:
    :Service:
        ``sarg.d``
    :Confiles:
        No
    :Variables:
        No

Fedora final configuration:
    :Replaces:
        Yes
    :Files:
        ``/etc/sarg/sarg.conf``

"""
import re
from gettext import gettext as _

from libmace.providers import BaseProvider
from libmace.util import enforce


_LANGS = {'bg': 'Bulgarian_windows1251',
          'ca': 'Catalan',
          'cs': 'Czech_UTF8',
          'nl': 'Dutch',
          'en': 'English',
          'fr': 'French',
          'de': 'German',
          'el': 'Greek',
          'hu': 'Hungarian',
          'id': 'Indonesian',
          'it': 'Italian',
          'jp': 'Japanese',
          'lv': 'Latvian',
          'pl': 'Polish',
          'pt': 'Portuguese',
          'ro': 'Romanian',
          'ru': 'Russian_UFT-8',
          'sr': 'Serbian',
          'sk': 'Slovak',
          'es': 'Spanish',
          'tk': 'Turkish',
          }


class Provider(BaseProvider):

    @property
    def comment(self):
        return '#'

    @property
    def name(self):
        return 'sarg'

    @property
    def depends(self):
        return ['httpd']

    def get_requires(self, conftext):
        return ['sarg']

    @property
    def service_name(self):
        return None

    @property
    def files(self):
        return [None]

    def confile(self, filename=None):
        return '/etc/sarg/sarg.conf'

    def convert(self, lines):
        for i, line in enumerate(lines):
            if line.startswith('language'):
                language = re.split('\s+', line, 1)[-1].strip()
                enforce(language in _LANGS, _('Cannot find language %s, ' \
                        'it should be in %r'), language, _LANGS.keys())
                lines[i] = 'language %s' % _LANGS[language]
        return lines

    def ensure(self, tran):
        output_dir = None

        for line in file(tran.path(self.confile())).readlines():
            if line.startswith('output_dir'):
                output_dir = re.split('\s+', line, 1)[-1].strip()

        enforce(output_dir,
                _('Option output_dir is not set in sarg configuration file'))
        tran.ensure_dir(output_dir, mode=0755)
