# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Squid
-----

Provider will do some extra work:

- make sure that swap directory is properly created
- make sure that log directories are properly created

If configuration template contains ``filter`` confile, squidQuard
will be enabled.

Mace configuration:
    :Service:
        ``squid.d``
    :Confiles:
        - ``access`` only for access rules
        - ``conf`` the rest of configuration
        - ``filter`` squidQuard configuration
    :Variables:
        No

Fedora final configuration:
    :Replaces:
        Yes
    :Files:
        - ``/etc/squid/squid.conf``
        - ``/etc/squid/squidGuard.conf``

"""
import logging
from os.path import dirname
from gettext import gettext as _

from libmace import util, env
from libmace.providers import BaseProvider


class Provider(BaseProvider):

    @property
    def comment(self):
        return '#'

    @property
    def name(self):
        return 'squid'

    def get_requires(self, conftext):
        requires = ['squid']
        if conftext.confile('filter'):
            requires.append('squidguard')
        return requires

    @property
    def service_name(self):
        return 'squid'

    @property
    def files(self):
        return ['access', 'conf', 'filter']

    def confile(self, filename=None):
        if filename == 'filter':
            return '/etc/squid/squidGuard.conf'
        else:
            return '/etc/squid/squid.conf'

    def confile_footer(self, filename, conftext):
        if conftext.confile('filter'):
            return ['url_rewrite_program /usr/bin/squidGuard -c %s' % \
                        self.confile('filter'),
                    ]
        else:
            return []

    def ensure(self, tran):
        self._ensure_conf_dirs(tran)
        # On Fedora, at least on Fedora-11,
        # init.d script uses hard coded log dir for its output
        tran.ensure_dir('/var/log/squid')

    def _ensure_conf_dirs(self, tran):
        user = 'squid'
        group = 'squid'
        swap_path = '/var/spool/squid'
        access_log = '/var/log/squid/access.log'
        cache_log = '/var/log/squid/cache.log'
        cache_store_log = '/var/log/squid/store.log'

        for line in file(tran.path(self.confile())).readlines():
            if line.startswith('cache_effective_user'):
                user = line.split()[-1]
            elif line.startswith('cache_effective_group'):
                group = line.split()[-1]
            elif line.startswith('cache_dir'):
                swap_path = line.split()[2]
            elif line.startswith('access_log'):
                access_log = line.split()[1]
            elif line.startswith('cache_log'):
                cache_log = line.split()[1]
            elif line.startswith('cache_store_log'):
                cache_store_log = line.split()[1]

        owner = (user, group)

        if tran.ensure_dir(swap_path, owner=owner):
            if not env.dry_services.value:
                logging.info(_('Create Squid swap'))
                util.assert_call(['squid', '-z'])

        for i in [access_log, cache_log, cache_store_log]:
            if i != 'none':
                tran.ensure_dir(dirname(i), owner=owner)
