# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Users and groups
----------------

Configure users and groups on final system.

This provider uses artificial Mace-level configuration syntax in ini-files
format. Each ini section should be unique and might configure user, user/group
or group depending on section keys.

Section keys are:
    :user:
        user name; if set user will be configured
    :group:
        group name; if set, group will be configured
    :groups:
        if user is being configured, this is semicolon sepparated list
        of groups
    :managehome:
        if not empty and user is being configured, create user's home directory
    :ssh-key:
        if user is being configured, the string that will be verbatimly added
        to `authorized_keys` file for this user

Mace configuration:
    :Service:
        ``users.d``
    :Confiles:
        No
    :Variables:
        No

Final configuration for all distributions:
    :Replaces:
        Yes
    :Files:
        - ``/etc/passwd``
        - ``/etc/group``
        - ``/etc/shadow``
        - ``/home/*/.ssh/authorized_keys``

"""
import pwd
import grp
from os.path import join

from libmace import distro
from libmace.providers import BaseProvider


class Provider(BaseProvider):

    @property
    def comment(self):
        return None

    @property
    def name(self):
        return 'users'

    @property
    def service_name(self):
        return None

    @property
    def files(self):
        return [None]

    def confile(self, filename):
        return None

    def apply(self, tran):
        stale = False
        conf = tran.config

        for section in conf.sections():
            group = conf[section, 'group']
            if group and not _group_exists(group):
                distro.group_add(group)
                stale = True

            user = conf[section, 'user']
            if user:
                home = '/home/%s' % user

                if not _user_exists(user):
                    kwargs = {}
                    if group:
                        kwargs['group'] = group
                    if conf[section, 'managehome']:
                        kwargs['managehome'] = True
                    if conf[section, 'groups']:
                        kwargs['groups'] = conf.get_list(section, 'groups')
                    distro.user_add(user, **kwargs)
                    stale = True
                else:
                    kwargs = {}
                    pwd_user = pwd.getpwnam(user)
                    if group and pwd_user.pw_gid != grp.getgrnam(group).gr_gid:
                        kwargs['group'] = group
                    if conf[section, 'managehome'] and pwd_user.pw_dir != home:
                        kwargs['home'] = home
                    if conf[section, 'groups']:
                        groups = [i.strip() for i in \
                                conf[section, 'groups'].split(',')]
                        if set(groups) != distro.get_user_groups(user):
                            kwargs['groups'] = groups
                    if kwargs:
                        distro.user_modify(user, **kwargs)
                        stale = True

                if conf[section, 'ssh-key']:
                    tran.ensure_dir(home,
                            owner=(user, group), mode=0755)
                    tran.ensure_dir(join(home, '.ssh'),
                            owner=(user, group), mode=0700)
                    tran.ensure_file(
                            join(home, '.ssh', 'authorized_keys'),
                            content='\n'.join(
                                conf.get_list(section, 'ssh-key')),
                            owner=(user, group), mode=0600)

        return stale or not tran.empty


def _group_exists(name):
    try:
        grp.getgrnam(name)
        return True
    except KeyError:
        return False


def _user_exists(name):
    try:
        pwd.getpwnam(name)
        return True
    except KeyError:
        return False
